/*
 *
 *
 *
 *
 */

#ifndef LIBPWC_H
#define LIBPWC_H

#include <linux/videodev.h>
#include <linux/videodev2.h>
#include "pwc-ioctl.h"

#define LIBPWC_VER_STRING "0.1.0"
#define LIBPWC_VER_MAJOR   0
#define LIBPWC_VER_MINOR   1
#define LIBPWC_VER_RELEASE 0
#define LIBPWC_VER_CODE	   ((LIBPWC_VER_MAJOR<<16)| \
                            (LIBPWC_VER_MINOR<<8)|  \
                            (LIBPWC_VER_RELEASE))

#define LIBPWC_VER_SONUM   0
#define LIBPWC_VER_DLLNUM  1

#ifndef V4L2_PIX_FMT_PWC1
# define V4L2_PIX_FMT_PWC1     v4l2_fourcc('P','W','C','1') /* pwc older webcam */                                
# define V4L2_PIX_FMT_PWC2     v4l2_fourcc('P','W','C','2') /* pwc newer webcam */
#endif

#ifndef DEVICE_USE_CODEC1
#define DEVICE_USE_CODEC1(x) ((x)<675)
#define DEVICE_USE_CODEC2(x) ((x)>=675 && (x)<700)
#define DEVICE_USE_CODEC3(x) ((x)>=700)
#define DEVICE_USE_CODEC23(x) ((x)>=675)
#endif


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

struct pwc_t {
   int type;			/* type of the webcam */
   unsigned int width, height;	/* can be filled by the user BEFORE calling libpwc_decode */
   unsigned char cmd[4];	/* command used to grab the image from the webcam */
   int vbandlength;		/* size of 4 compressed row */
   int codec_version;
   void *tables;		/* Dynamic lookup table */
};

struct pwc_t *libpwc_init(void);
int libpwc_open(struct pwc_t *pwc, const unsigned char *devname);
int libpwc_close(struct pwc_t *pwc);
int libpwc_release(struct pwc_t *pwc);
int libpwc_decode(struct pwc_t *pwc, 
     const int raw_format, const unsigned char *rawstream, 
     const int out_format, void **buffer);

int libpwc_decode_simple(int width, int height, const void *rawstream, int out_format, void **buffer);

/* Function to convert data */
void pwc_yuv420p_to_rgb(unsigned int w, unsigned int h, 
    			const unsigned char *py,
			const unsigned char *pu,
			const unsigned char *pv,
			unsigned char *d);


#ifdef __cplusplus
}
#endif

#endif

